;; Generated package description from jsonrpc.el  -*- no-byte-compile: t -*-
(define-package "jsonrpc" "1.0.14.0.20220113.222" "JSON-RPC library" '((emacs "25.2")) :commit "626c1dce02d4f2e8d0822decf8b421eee00674b1" :url "https://elpa.gnu.org/packages/jsonrpc.html" :authors '(("João Távora" . "joaotavora@gmail.com")) :maintainer '("João Távora" . "joaotavora@gmail.com") :keywords '("processes" "languages" "extensions"))
